@@ -109,10 +109,8 @@ class BaseConnection(Generic[Row]):
         self._notice_handlers: List[NoticeHandler] = []
         self._notify_handlers: List[NotifyHandler] = []
 
-        # Stack of savepoint names managed by current transaction blocks.
-        # the first item is "" in case the outermost Transaction must manage
-        # only a begin/commit and not a savepoint.
-        self._savepoints: List[str] = []
+        # Number of transaction blocks currently entered
+        self._num_transactions = 0
 
         self._closed = False  # closed by an explicit close()
         self._prepared: PrepareManager = PrepareManager()
@@ -247,7 +245,7 @@ class BaseConnection(Generic[Row]):
         # Raise an exception if we are in a transaction
         status = self.pgconn.transaction_status
         if status != TransactionStatus.IDLE:
-            if self._savepoints:
+            if self._num_transactions:
                 raise e.ProgrammingError(
                     f"can't change {attribute!r} now: "
                     "connection.transaction() context in progress"
@@ -404,13 +402,7 @@ class BaseConnection(Generic[Row]):
         Only used to implement internal commands such as "commit", with eventual
         arguments bound client-side. The cursor can do more complex stuff.
         """
-        if self.pgconn.status != ConnStatus.OK:
-            if self.pgconn.status == ConnStatus.BAD:
-                raise e.OperationalError("the connection is closed")
-            raise e.InterfaceError(
-                f"cannot execute operations: the connection is"
-                f" in status {self.pgconn.status}"
-            )
+        self._check_connection_ok()
 
         if isinstance(command, str):
             command = command.encode(pgconn_encoding(self.pgconn))
@@ -437,6 +429,17 @@ class BaseConnection(Generic[Row]):
                 )
         return result
 
+    def _check_connection_ok(self) -> None:
+        if self.pgconn.status == ConnStatus.OK:
+            return
+
+        if self.pgconn.status == ConnStatus.BAD:
+            raise e.OperationalError("the connection is closed")
+        raise e.InterfaceError(
+            f"cannot execute operations: the connection is"
+            f" in status {self.pgconn.status}"
+        )
+
     def _start_query(self) -> PQGen[None]:
         """Generator to start a transaction if necessary."""
         if self._autocommit:
@@ -471,7 +474,7 @@ class BaseConnection(Generic[Row]):
 
     def _commit_gen(self) -> PQGen[None]:
         """Generator implementing `Connection.commit()`."""
-        if self._savepoints:
+        if self._num_transactions:
             raise e.ProgrammingError(
                 "Explicit commit() forbidden within a Transaction "
                 "context. (Transaction will be automatically committed "
@@ -484,7 +487,7 @@ class BaseConnection(Generic[Row]):
 
     def _rollback_gen(self) -> PQGen[None]:
         """Generator implementing `Connection.rollback()`."""
-        if self._savepoints:
+        if self._num_transactions:
             raise e.ProgrammingError(
                 "Explicit rollback() forbidden within a Transaction "
                 "context. (Either raise Rollback() or allow "
@@ -677,6 +680,8 @@ class Connection(BaseConnection[Row]):
         """
         Return a new cursor to send commands and queries to the connection.
         """
+        self._check_connection_ok()
+
         if not row_factory:
             row_factory = self.row_factory
 
@@ -706,12 +711,13 @@ class Connection(BaseConnection[Row]):
         binary: bool = False,
     ) -> Cursor[Row]:
         """Execute a query and return a cursor to read its results."""
-        cur = self.cursor()
-        if binary:
-            cur.format = Format.BINARY
-
         try:
+            cur = self.cursor()
+            if binary:
+                cur.format = Format.BINARY
+
             return cur.execute(query, params, prepare=prepare)
+
         except e.Error as ex:
             raise ex.with_traceback(None)
 
