@@ -6,7 +6,7 @@ psycopg cursor objects
 
 import sys
 from types import TracebackType
-from typing import Any, Callable, Generic, Iterable, Iterator, List
+from typing import Any, Generic, Iterable, Iterator, List
 from typing import Optional, NoReturn, Sequence, Type, TypeVar, TYPE_CHECKING
 from contextlib import contextmanager
 
@@ -29,8 +29,6 @@ if TYPE_CHECKING:
     from .pq.abc import PGconn, PGresult
     from .connection import Connection
 
-execute: Callable[["PGconn"], PQGen[List["PGresult"]]]
-
 if _psycopg:
     execute = _psycopg.execute
     fetch = _psycopg.fetch
@@ -53,7 +51,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
         __slots__ = """
             _conn format _adapters arraysize _closed _results pgresult _pos
             _iresult _rowcount _query _tx _last_query _row_factory _make_row
-            _pgconn
+            _pgconn _encoding
             __weakref__
             """.split()
 
@@ -80,6 +78,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
         self._iresult = 0
         self._rowcount = -1
         self._query: Optional[PostgresQuery]
+        self._encoding = "utf-8"
         if reset_query:
             self._query = None
 
@@ -157,14 +156,8 @@ class BaseCursor(Generic[ConnectionType, Row]):
         Return `!True` if a new result is available, which will be the one
         methods `!fetch*()` will operate on.
         """
-        self._iresult += 1
-        if self._iresult < len(self._results):
-            self.pgresult = self._results[self._iresult]
-            self._tx.set_pgresult(self._results[self._iresult])
-            self._make_row = self._make_row_maker()
-            self._pos = 0
-            nrows = self.pgresult.command_tuples
-            self._rowcount = nrows if nrows is not None else -1
+        if self._iresult < len(self._results) - 1:
+            self._set_result(self._iresult + 1)
             return True
         else:
             return None
@@ -204,7 +197,8 @@ class BaseCursor(Generic[ConnectionType, Row]):
         results = yield from self._maybe_prepare_gen(
             pgq, prepare=prepare, binary=binary
         )
-        self._execute_results(results)
+        self._set_results(results)
+        self._set_result(0)
         self._last_query = query
 
     def _executemany_gen(
@@ -213,6 +207,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
         """Generator implementing `Cursor.executemany()`."""
         yield from self._start_query(query)
         first = True
+        nrows = 0
         for params in params_seq:
             if first:
                 pgq = self._convert_query(query, params)
@@ -222,8 +217,16 @@ class BaseCursor(Generic[ConnectionType, Row]):
                 pgq.dump(params)
 
             results = yield from self._maybe_prepare_gen(pgq, prepare=True)
-            self._execute_results(results)
+            self._set_results(results)
 
+            for res in results:
+                nrows += res.command_tuples or 0
+
+        self._set_result(0)
+        # Override rowcout for the first result. Calls to nextset() will change
+        # it to the value of that result only, but we hope nobody will notice.
+        # You haven't read this comment.
+        self._rowcount = nrows or -1
         self._last_query = query
 
     def _maybe_prepare_gen(
@@ -232,7 +235,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
         *,
         prepare: Optional[bool] = None,
         binary: Optional[bool] = None,
-    ) -> PQGen[Sequence["PGresult"]]:
+    ) -> PQGen[List["PGresult"]]:
         # Check if the query is prepared or needs preparing
         prep, name = self._conn._prepared.get(pgq, prepare)
         if prep is Prepare.YES:
@@ -248,9 +251,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
             self._send_prepare(name, pgq)
             (result,) = yield from execute(self._pgconn)
             if result.status == ExecStatus.FATAL_ERROR:
-                raise e.error_from_result(
-                    result, encoding=pgconn_encoding(self._pgconn)
-                )
+                raise e.error_from_result(result, encoding=self._encoding)
             self._send_query_prepared(name, pgq, binary=binary)
 
         # run the query
@@ -316,6 +317,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
             raise e.InterfaceError("the cursor is closed")
 
         self._reset()
+        self._encoding = pgconn_encoding(self._pgconn)
         if not self._last_query or (self._last_query is not query):
             self._last_query = None
             self._tx = adapt.Transformer(self)
@@ -387,13 +389,9 @@ class BaseCursor(Generic[ConnectionType, Row]):
         ExecStatus.COPY_BOTH,
     )
 
-    def _execute_results(
-        self, results: Sequence["PGresult"], format: Optional[Format] = None
-    ) -> None:
+    def _set_results(self, results: List["PGresult"]) -> None:
         """
-        Implement part of execute() after waiting common to sync and async
-
-        This is not a generator, but a normal non-blocking function.
+        Set the results from a query into the cursor state.
         """
         if not results:
             raise e.InternalError("got no result from the query")
@@ -402,29 +400,30 @@ class BaseCursor(Generic[ConnectionType, Row]):
             if res.status not in self._status_ok:
                 self._raise_from_results(results)
 
-        self._results = list(results)
-        self.pgresult = results[0]
+        self._results = results
+
+    def _set_result(self, i: int, format: Optional[Format] = None) -> None:
+        """
+        Select one of the results in the cursor as the active one.
+        """
+        self._iresult = i
+        res = self.pgresult = self._results[i]
 
         # Note: the only reason to override format is to correclty set
         # binary loaders on server-side cursors, because send_describe_portal
         # only returns a text result.
-        self._tx.set_pgresult(results[0], format=format)
+        self._tx.set_pgresult(res, format=format)
 
         self._make_row = self._make_row_maker()
+        self._pos = 0
         nrows = self.pgresult.command_tuples
-        if nrows is not None:
-            if self._rowcount < 0:
-                self._rowcount = nrows
-            else:
-                self._rowcount += nrows
+        self._rowcount = nrows if nrows is not None else -1
 
-    def _raise_from_results(self, results: Sequence["PGresult"]) -> NoReturn:
+    def _raise_from_results(self, results: List["PGresult"]) -> NoReturn:
         statuses = {res.status for res in results}
         badstats = statuses.difference(self._status_ok)
         if results[-1].status == ExecStatus.FATAL_ERROR:
-            raise e.error_from_result(
-                results[-1], encoding=pgconn_encoding(self._pgconn)
-            )
+            raise e.error_from_result(results[-1], encoding=self._encoding)
         elif statuses.intersection(self._status_copy):
             raise e.ProgrammingError(
                 "COPY cannot be used with this method; use copy() insead"
@@ -469,9 +468,7 @@ class BaseCursor(Generic[ConnectionType, Row]):
         if status in (ExecStatus.COPY_IN, ExecStatus.COPY_OUT):
             return
         elif status == ExecStatus.FATAL_ERROR:
-            raise e.error_from_result(
-                result, encoding=pgconn_encoding(self._pgconn)
-            )
+            raise e.error_from_result(result, encoding=self._encoding)
         else:
             raise e.ProgrammingError(
                 "copy() should be used only with COPY ... TO STDOUT or COPY ..."
